/**
 * Dynamic Field Theory has at its core a convolution between a kernel, 
 * usually a gaussian-type function, and a vector representing the field
 * of interest, more specifically it is a sigmoidal filtered version of 
 * a the vector that represents a field, Eq. \ref equ_getConv1D "Conv1D".
 * BADdft2 has been implemented using the Integrated Performance 
 * Primitives (IPP) libraries for Intel® architectures. These libraries
 * allow BADdft2 to improve the speed of the convolutions required by
 * each field.
 * 
 * This library gives the user the possibility to create networks of 1D-,
 * 2D, and 3D-fields, and any possible interaction among them. In the 
 * present revision, the 2D- and 3D- fields are computed using one-dimensional
 * convolutions.
 * 
 * \author 		Boris Duran
 * \date   		2010-2011
 * \bug				None detected :) I'll fix them as soon as you find one.
 * \warning   3D fields can still take some time depending on the sizes
 * 						of each dimension.
 * \copyright	GNU Library or Lesser General Public License (LGPL)
 *
 */

#ifndef BADDFT2_H
#define BADDFT2_H

# ifdef _WIN32
    #define BDLL __declspec(dllexport)
# else
    #define BDLL
# endif

// ----- STD ----
#include <vector>
#include <cmath>
#include <iostream>
#include <cstdlib>

// ----- IPP ----
#include "ippdefs.h"
#include "ippcore.h"
#include "ippi.h"
#include "ipps.h"
#include "ippvm.h"

// --- MyLibs ---
#include "badtools.h"

namespace BAD{

	///< \struct Dimension
	///< \brief Dimensions parameters
	/** 
	 * Contains specific information about each of the dimensions of a field. 
	 */
	struct Dimension{
		std::string name;		///< Name of this dimension.
		int sizeD;				///< Size of a dimension.
		int dMin;				///< Minimum value for this dimension.
		int dMax;				///< Maximum value for this dimension.
		
		public:
			///< \brief Default Dimension constructor
			Dimension(){}
			///< \brief Default Dimension destructor
			~Dimension(){ name.clear(); }
	};

	///< \struct Kernel
	///< \brief Kernel parameters and data vectors.
	/** 
	 * Defines all parameters necessary for the creation of a 
	 * mexican-hat type of kernel with global inhibition. 
	 */
	struct Kernel{
		/** Defines if a field will inhibit(-1) or excite(+1) the one that 
			 it is connected to.*/
		Ipp64f type;		///< Sign of connectivity. 
		int idOrig;			///< ID of Field where the output is coming from.
		int idDimS;			///< ID of Dimension in source field.
		int idDimT;			///< ID of Dimension in target field.
		int sizeK[2];		///< Size of this kernel.
		Ipp64f* dataKe;		///< The actual kernel.
		Ipp64f* dataKi;		///< The actual kernel.

		Ipp64f c_exc;		///< Kernel's excitatory amplitude.
		Ipp64f s_exc;		///< Kernel's excitatory width.
		Ipp64f c_inh;		///< Kernel's inhibitory amplitude.
		Ipp64f s_inh;		///< Kernel's inhibitory width.
		Ipp64f g_inh;		///< Global inhibition.

		bool is1D;			///< Identifies a single dimensional field

		public:
			///< \brief Default constructor for kernel structures.
			/**
			* Initializes all parameters to default values.
			*/
			Kernel(){
//				std::cout << std::endl << "===>>> Empty Kernel constructor! <<<===";
				type = 1.0;				idOrig = 0;
				sizeK[0] = 1;			sizeK[1] = 1;
				idDimS = 0;				idDimT = 0;
				
				c_exc = 0.0;	s_exc = 0.0;
				c_inh = 0.0;	s_inh = 0.0;	g_inh = 0.0;
				
				dataKe = NULL;
				dataKi = NULL;
			}

			///< \brief Copy constructor for kernel structures.
			/**
			* Sets all parameters of the new kernel to the values obtained
			* from the referenced kernel structure and allocates memory for 
			* the data of the kernel itself both in its excitatory and
			* inhibitory parts.
			* @param kernel Pointer to the kernel structure to be copied.
			*/
			Kernel(const Kernel& kernel){
//				std::cout << std::endl << "===>>> Copy kernel constructor! <<<==="
//									<< std::endl;
				type = kernel.type;
				idOrig = kernel.idOrig;
				idDimS = kernel.idDimS;
				idDimT = kernel.idDimT;
				sizeK[0] = kernel.sizeK[0];
				sizeK[1] = kernel.sizeK[1];
				
				c_exc = kernel.c_exc;		s_exc = kernel.s_exc;
				c_inh = kernel.c_inh;		s_inh = kernel.s_inh;
				g_inh = kernel.g_inh;
				
				dataKe = NULL;
				dataKi = NULL;
			}

			///< \brief Destructor for kernel structures.
			~Kernel(){
				//std::cout << ">>> Destroying Kernel structure... ";
				ippsFree( dataKe );
				ippsFree( dataKi );
				//std::cout << "done destroying kernel. <<<" << std::endl;
			}
	};
	
	///< \struct Field
	///< \brief Field parameters and data vectors.
	/**
	* Defines all parameters necessary for the activation
	* of a N-dimensional dynamic field.
	*/
	struct Field {

		std::string name;		///< Field's name given by the designer.
		int id;					///< The identifier of this field.
		int nDim;				///< Number of dimensions for this field.
		int sizeF;				///< Total size of the field.
		int nKernels;			///< Number of kernels working with this field.
		int nInteractions;		///< Number of interactions with this field.

		Ipp64f tau_build;		///< Field's integration time (a peak is built).
		Ipp64f tau_decay;		///< Field's integration time (a peak decays).
		Ipp64f tau_peak;		///< Integration time for peak detection (default = 10.)
		Ipp64f h;				///< Field's resting level.
		Ipp64f beta;			///< Steepness of the sigmoid filter.
		Ipp64f c_noise;			///< Amplitude of the field's internal noise.
		Ipp64f s_noise;			///< Width of the field's internal noise.

		Ipp64f* data;			///< The actual field.
		BADtools *tools;		///< A badTools object.
		std::vector<Kernel> kernelList;	///< All interactions with this field
		std::vector<Dimension> dimList;	///< Info about each dimension.
		
		int sizeN;				///< Size of the noise kernel.
		Ipp64f* kernelN;		///< Saves the data of the kernel for noise.

		public:
			///< \brief Default constructor for field structures.
			/**
			* Initializes all parameters to default values.
			*/
			Field(){
//				std::cout << "===>>> Empty Field constructor! <<<===" << std::endl;
				name = "1D";	id = 0;	nDim = 1;
				sizeF = 0;	nKernels = 0;	nInteractions = 0;	sizeN = 0;
				
				tau_build = 0.0;	tau_decay = 0.0;	tau_peak = 0.0;
				h = 0.0;	beta = 0.0;	c_noise = 0.0;	s_noise = 0.0;
				
				dimList.clear();
				kernelList.clear();
				tools = NULL;
				kernelN = NULL;
				data = NULL;
			}
			
			///< \brief Copy constructor for field structures.
			/**
			* Sets all parameters of the new field to the values obtained
			* from the referenced field structure and allocates memory for
			* the data of the field itself.
			* @param field Pointer to the field structure to be copied.
			*/
			Field(const Field& field){
//				std::cout << "===>>> Copy Field constructor! <<<===" << std::endl;
				name = field.name;	
				id = field.id;	
				nDim = field.nDim;
				sizeF = field.sizeF;
				sizeN = field.sizeN;
				nKernels = field.nKernels;	
				nInteractions = field.nInteractions;
				
				tau_build = field.tau_build;	
				tau_decay = field.tau_decay;	
				tau_peak = field.tau_peak;
				h = field.h;	
				beta = field.beta;	
				c_noise = field.c_noise;	
				s_noise = field.s_noise;
				
				dimList = field.dimList;
				kernelList = field.kernelList;
				tools = NULL;
/*
				if( nDim == 1 )
					tools = new BADtools( dimList[0].dMin, dimList[0].dMax );
				else if(nDim == 2 )
					tools = new BADtools( dimList[0].dMin, dimList[0].dMax,
                                dimList[1].dMin, dimList[1].dMax );
*/
				data = NULL;
				kernelN = NULL;
			}

			///< \brief Destructor for field structures.
			~Field(){
				//std::cout << ">>> Destroying Field structure... " << std::endl;
				name.clear();
				ippsFree( data );
				ippsFree( kernelN );
				dimList.clear();
				kernelList.clear();
				delete tools;
				//std::cout << "done destroying field. <<<" << std::endl;
			}
	};

	///< \struct Sequence
	///< \brief Sequence parameters
	/**
	* Defines all parameters necessary for the activation
	* of a sequence of neural activations.
	*/
	struct Sequence {
		Ipp64f tau;			///< Node's integration time.
		Ipp64f h;				///< Node's resting level.
		Ipp64f beta;		///< Steepness of the sigmoidal filter.
		Ipp64f c_self;	///< Gain from a node into itself.
		Ipp64f c_xtrn1;	///< Gain from an external node (current node in the sequence).
		Ipp64f c_xtrn2;	///< Gain from an external node (previous node in the sequence).
		Ipp64f c_noise;	///< Amplitude of the sequence internal noise.
		Ipp64f g_inh;		///< Global inhibition.
	};

	///< \class badDFT.
	///< \brief Dynamic neural fields.
	/** 
	 * This library gives the user the possibility to create networks of
	 * 1D-, 2D-, and 3D-fields, and any possible interaction among them.
	 * In the present revision, the 2D- and 3D- fields are computed using
	 * one-dimensional convolutions. Performance increases a little bit 
	 * this way compared to a 2D convolution.
	 * 
	 * @todo Interaction between 1D<->3D fields and 2D<->3D fields.
	 * @todo Check that the 2D convolution can be circular.
	 */
	class badDFT2{

		private:
			Sequence *pSS;
			int kFactor, maxSize;
			bool verbose, hasConfig, isCircular;
			Ipp64f *seqOrdinal, *seqMemory;
			
		public:
			int tMax;		///< Number of timesteps the simulation will run.
			int nFields;	///< Total number of fields present in the network.
			int sSequence;	///< Size of the sequence.
			Ipp64f tau_peak;///< Integration time for peak detection (default = 10.)
			Ipp64f *bufferL;///< A working buffer.
			std::vector<Field> fieldList;	///< A vector of all fields in this network.
			float tSum;
			
			BDLL badDFT2();		///< \brief Default Constructor.

			///< \brief Constructor with configuration file.
			/**
			* Sets all parameters and allocate memory for all kernels and 
			* fields. It finally initializes the range with which all fields
			* will be working with, initializes all kernels and sets up the
			* resting level to all fields.
			*/
			BDLL badDFT2( const char *cfgName );	

			/**
			* Frees memory.
			*/
			BDLL virtual ~badDFT2();				///< \brief Destructor.

			/** This function initializes all parameters needed by a 
			 * dynamic field network to work properly. 
			 * @param cfgName Name of the external configuration file.
			 * @return Updates the different parameters.\n
			 */
			int setParams( const char *cfgName );	///< \brief	Sets parameters. 

			/** Allocates memory for all kernels and fields.
			 */
			void allocateMemory();	///< \brief	Allocates memory. 
			
			/** Initializes all kernels and fields.
			 * @return Updates the activation of kernels and fields.\n
			 */
			BDLL void initAll();		///< \brief	Resets fields and kernels. 
			
			/** \brief	Generates the data for kernels.*/
			/** Generates the kernels to be used in the convolution with fields.
			 * @param fOrigin The field where the interaction originates.
			 * @param fTarget The field where the	interaction finishes.
			 * @param kDim Index of the current dimension.
			 * @param kCount Index of the current kernel.
			 * @return Updates the value of dataKe and dataKi.\n
			 */
			void getKernel( Field *fOrigin, Field *fTarget, int kDim, int kCount );

			/**< \brief Performs a single step.*/
			/** Contains a loop through all fields which are ...
			 * The standard Amari dynamics is being used:
			 * \f{equation}{ \tau \dot{u}_W(x,t) = -u_W(x,t) + h + S(x,t) + 
			 * 							 \int{w(x-x')f[u_W(x',t)]}dx' \f}
			 * Where 'f' represents the sigmoid function in this library.			 */
			BDLL void dftStep( std::vector<Ipp64f*> inputs );

    /** \brief	Updates the sequence nodes.*/
    /** Gets the activation of all ordinal and memory nodes in the
     * predefined sequence. The following equation defines the dynamics
     * of both ordinal and memory nodes:
     * \f{equation}{ 
     *		\tau \dot{d}_i(t) =  -d_i(t) + h_d + c_0 f[d_i(t)] 
     * 			- c_1 \sum_{i\prime \neq i} f[d_i\prime(t)] 
     * 			+ c_2 f[d_{i-1}^m(t)] - c_3 f[d_i^m(t)] - I_c(t).
     * \f}
     * \f{equation}{ 
     *		\tau \dot{d}_i^m(t) =  -d_i^m(t) + h_m + c_4 f[d_i^m(t)] 
     * 			- c_5 \sum_{i\prime \neq i} f[d_i\prime(t)] + c_6 f[d_i(t)].
     * \f}
     * @param fCoS Pointer to a field structure that represents the 
     * Condition of Satisfaction field especified in the configuration 
     * file.
     * @param go_signal Starts (go_signal > 0) the sequence. This value
     * should be set to zero after a few timesteps, and big enough so 
     * that the action field builds a peak.
     * @return Updates the activation of bufferS.\n
     */
    BDLL void seqStep( Field *fCoS, int go_signal=0, Ipp32f *weights=NULL );
    
			/** \brief	Updates field activations in 1D using Convolution (IPP).*/
    /** \anchor equ_getConv1D
     * Gets the interaction between a field and a kernel by applying
     * the convolution between those components. The convolution 
     * is computed using the IPP libraries.
     * 
     * \f{equation}{ \int{w(x-x')f[u_W(x',t)]}dx' => 
     * 								conv(w(x-x'), f(u_W(x',t)) )  
     * 							 
     * \f}
     * Where 'f' represents the sigmoid function in this library.
     * @param fOrigin The field where the interaction originates.
     * @param kernel The kernel describing the interaction between both fields.
     * @param fTarget The field where the	interaction finishes.
     * @return Updates the value of bufferL.\n
     */
			void getConv1D( Field *fOrigin, Kernel *kernel, Field *fTarget );

			/** \brief	Updates field activations in 2D using Convolution (IPP).*/
			/** Gets the convolution of a vector representing a two-dimensional
			 * field given by "field" and two vectors representing the respective
			 * kernels for the convolution in both dimensions. The convolution 
			 * is computed using the IPPi  libraries.
			 * @param field A vector representing a 2D matrix.
			 * @param kernelX A kernel used for the convolution in dimension X.
			 * @param kernelY A kernel used for the convolution in dimension Y.
			 * @return Updates the activation of bufferL.\n
			 */
			void getConv2D( Field *field, Kernel *kernelX, Kernel *kernelY );

			/** \brief	Updates field activations in 3D (IPP).*/
			/** Gets the convolution of a vector representing a three-dimensional
			 * field given by "field" and three vectors representing the
			 * respective kernels for the convolution in each dimension. 
			 * @param field A vector representing a 3D matrix.
			 * @param kernelX A 1D kernel used for the convolution in dimension X.
			 * @param kernelY A 1D kernel used for the convolution in dimension Y.
			 * @param kernelZ A 1D kernel used for the convolution in dimension Z.
			 * @return Updates the activation of bufferL.\n
			 */
			void getConv3D( Field *field, Kernel *kernelX, 
													Kernel *kernelY, Kernel *kernelZ );

			/** \brief	Gets the 1D component of an interaction with a 2D field.*/
    /** Gets the activation of a two-dimensional fields and integrates
     * over the common dimension with the one-dimensional vector to which
     * the interaction is required. 
     * @param fOrigin The origin vector representing a 2D matrix.
     * @param kernel 	A kernel used for the convolution in dimension X.
     * @param fTarget The target vector representing a 1D array.
     * @return Updates the activation of bufferL.\n
     */
			void getConv2Dto1D( Field *fOrigin, Kernel *kernel, Field *fTarget );

			/** \brief	Updates memory field.*/
			/** Gets the activation of a memory field connected to a single
			 * dynamic field. The following equation defines the dynamics of
			 * this type of field:
			 * \f{equation}{ 
			 *		\tau_m \dot{u}_M(x,t) =  \left[ -u_M(x,t) + f[u_W(x',t)] \right]   
			 *			\times \left[ \lambda_{build} \Theta(u_W(x,t)) + \lambda_{decay} (1 - \Theta(u_W(x,t))) \right]
			 * \f}
			 * @param ixMemory Index of the memory field.
			 * @param ixField Index of the field to be memorized.
			 * @return Updates the activation of bufferS.\n
			 */
			void getMemory( int ixMemory, int ixField );

			/** \brief	Adds Gaussian noise to a dynamic field.*/
			/** Adds Gaussian noise of mean=0 and standard deviation = 1 to 
			 * bufferS. If the parameter s_noise of a field is larger than 
			 * zero then a convolution is performed between the default noise
			 * vector with a kernel of amplitude 1 and s_noise-width.
			 * @param ixField Index of the field where noise will be added.
			 * @return Updates the activation of bufferS.\n
			 */
			void getNoise( int ixField );

			/** \brief	Gets the peak of a dynamic field.*/
			/** Obtains the position in the field where a peak has been 
			 * formed. This is done dynamically by integrating the field over
			 * its range (an integration time must be specified in the 
			 * configuration file):
			 * \f{equation}{ 
				  \tau_p \dot{x} = -x\int{f[u_W(x',t)]}dx' + \int{x' f[u_W(x',t)]}dx'
				 \f}
			 * Where 'f' represents the sigmoid function in this library.
			 * @param idField Index of the field of interest.
			 * @param peakPos Pointer to the output result.
			 * @return Updates the value representing the peak of a field.\n
			 */
			void getPeak1D( int idField, Ipp64f* peakPos );

			/** Saves all kernels into a text file.
			 * @param name Name of the file where kernels will be saved..
			 * @return True if data was saved, False otherwise.\n
			 */
			BDLL bool saveKernels( const char *name );		///< \brief	Saves Kernels. 

			/** Convolution with a 1D kernel for multi-dimensional fields.
			 * @param data The n-dimensional matrix reshaped to 1D vector.
			 * @param sDim1 Depth
			 * @param sDim2 Width
			 * @param sDim3 Height
			 * @param kernel Kernel for the width dimension
			 * @param sKernel Size of kernel
			 * @return Result of convolution in data.\n
			 */
      void convMD( Ipp64f *data, int sDim1, int sDim2, int sDim3, 
                  Ipp64f *kernel, int sKernel );

			/** Convolution with a 1D kernel for 2D-fields.
			 * @param data The 2D matrix reshaped to 1D vector.
			 * @param sDim1 Width
			 * @param sDim2 Depth
			 * @param kernel Kernel for the width dimension
			 * @param sKernel Size of kernel
			 * @return Result of convolution in data.\n
			 */
      void conv2D( Ipp64f *data, int sDim1, int sDim2, 
                  Ipp64f *kernel, int sKernel );

	};
}

#endif  // BADDFT2_H

