/**
 * 
 * A library for generic functions needed when working with dynamic field
 * theory. This library is based on the Intel® Integrated Performance 
 * Primitives (IPP) for Intel® Architecture. It is a collection of basic
 * mathematic functions such as the sigmoid and gaussian filters for one-
 * two- and three dimensions. There is also a function for shifting 
 * dimensions when working with fields of 3+ dimensions similar to what
 * Matlab's 'shiftdim' does.
 * 
 * \author 		Boris duran
 * \date   		2010-2011
 * \bug				None detected :) I'll fix them as soon as you find one.
 * \warning   Be careful about the limits of your kernels and fields. 
 * 						The size of thsoe vectors must be consistent with the
 * 						difference between the limits you specify.
 */

#ifndef BADTOOLS_H
#define BADTOOLS_H

# ifdef _WIN32
    #define BDLL __declspec(dllexport)
# else
    #define BDLL
# endif

//#include "ippdefs.h"
#include "ipps.h"
#include "ippi.h"
#include "ippvm.h"

namespace BAD{
	///< \class BADtools.
	///< \brief Generic functions for dynamic fields.
	/**
	 * A library for generic functions needed when working with dynamic field
	 * theory. This library is based on the Intel® Integrated Performance 
	 * Primitives (IPP) for Intel® Architecture. It is a collection of basic
	 * mathematic functions such as the sigmoid and gaussian filters for one-
	 * two- and three dimensions. There is also a function for shifting 
	 * dimensions when working with fields of 3+ dimensions similar to what
	 * Matlab's 'shiftdim' does.
	 */
	class BADtools{

		private:
			int sFieldX, sFieldY, sFieldZ;		// Size of fields
			bool hasParameters;

		public:
			int sMax;					///< Max size among all dimensions.
			Ipp64f *rangeIx;	///< Working range for fields in X.
			Ipp64f *rangeIy;	///< Working range for fields in Y.
			Ipp64f *rangeIz;	///< Working range for fields in Z.
			Ipp64f *bufferS;	///< Working vector of size sFieldX.
			Ipp64f *bufferM;	///< Working vector of size sFieldX*sFieldY.
			Ipp64f *bufferL;	///< Working vector of size sFieldX*sFieldY*sFieldZ.
			
			BADtools();			///< \brief Default Constructor.

			///< \brief Constructor for vectors.
			/**
			* Sets sizes of vectors and allocate memory for buffers and 
			* ranges. It initializes range vectors for building kernels
			* and inputs. Kernels are forced to be symmetric [-size/2:size/2]
			* whereas vectors themselves can be asymmetric [0:size).
			* @param xMin Minimum value for X dimension.
			* @param xMax Maximum value for X dimension.
			* @param yMin Minimum value for Y dimension.
			* @param yMax Maximum value for Y dimension.
			* @param zMin Minimum value for Z dimension.
			* @param zMax Maximum value for Z dimension.
			*/
			BDLL BADtools( int minx, int maxX, int minY=0, int maxY=1,
								int minZ=0, int maxZ=1);

			/**
			* Frees memory.
			*/
			BDLL virtual ~BADtools();				///< \brief Destructor.

			/** \brief	Shift dimensions.*/
			/** A hand-made function for shifting dimensions in vectors that represent
			 * 3D matrices. This function does the same as Matlab's 'shiftdim'.
			 * @param src Pointer to the vector being shifted.
			 * @param sDim1 First dimension to be shifted
			 * @param sDim2 Second dimension to be shifted
			 * @param sDim3 Third dimension to be shifted
			 * @param dst	Pointer to the vector where the shifted 'src' will be saved
			 * @return Updates the content of 'dst'.\n
			 */
			BDLL void shiftdim( Ipp64f *src, int sDim1, int sDim2, int sDim3, Ipp64f *dst );

			/** \brief	Sigmoidal function.*/
			/** A generic form of a sigmoidal filter.
			 * This functions is used to filter the output of a field.
			 * \f{equation}{ f(x) = \frac{1}{1+exp[-\beta(x-x_0)]} \f}
			 * @param field Vector to be filtered.
			 * @param beta Steepness of the sigmoid.
			 * @param x0 Bias of the filter.
			 * @param size Size of the vector to be filtered.
			 * @return Updates the content of bufferM.\n
			 */
			BDLL void sigmoid( Ipp64f *field, Ipp64f beta, Ipp64f x0, int size );

			/** \brief	1D Gaussian function.*/
			/** Applies a gaussian filter to a 1D vector.
			 * This functions is used to create external stimuli in 1D.
			 * \f{equation}{ f(x) = C\;exp\left[ \frac{-(x-x0)^2}{2\sigma^2} \right] \f}
			 * @param gain Amplitude for the gaussian function.
			 * @param x0 Location of the peak (mean).
			 * @param sigma Width of the distribution (variance).
			 * @param size Size of the vector to be filtered.
			 * @return Updates the content of bufferS.\n
			 */
			BDLL void gauss1D( Ipp64f gain, Ipp64f x0, Ipp64f sigma, int size );

			/** \brief	Normalized 1D Gaussian function.*/
			/** Gauss1Dnorm Function. 
			 * Applies a gaussian filter to a 1D vector and normalizes it.
			 * This functions is used to create kernels.
			 * \f{equation}{ 
				 * f(x) = exp\left[ \frac{-(x-x0)^2}{2\sigma^2} \right]; \quad 
				 * f(x) = C\;\frac{f(x)}{\sum{f(x)}}
			 * \f}
			 * @param gain Amplitude for the gaussian function.
			 * @param x0 Location of the peak (mean).
			 * @param sigma Width of the distribution (variance).
			 * @param size Size of the half vector to be filtered.
			 * @return Updates the content of bufferS.\n
			 */
			BDLL void gauss1Dnorm( Ipp64f gain, Ipp64f x0, Ipp64f sigma, int size );

			/** \brief	2D Gaussian function.*/
			/** Gauss2D Function. 
			 * Generates gaussian distribution in two dimensions with its 
			 * maximum at (x0,y0) controlled by 'gain' and variance 
			 * by 'sigmaX' and 'sigmaY' respectively.
			 * This functions is used to create external stimuli in 2D.
			 * \f{equation}{ 
				 * f(x,y) = C\;exp\left[ \frac{-(x-x0)^2}{2\sigma_x^2} +
				 * \frac{-(y-y0)^2}{2\sigma_y^2} \right] 
			 * \f}
			 * @param gain Amplitude for the gaussian function.
			 * @param x0 Location of the peak in X (mean).
			 * @param sigmaX Width of the distribution in X (variance).
			 * @param y0 Location of the peak in Y (mean).
			 * @param sigmaY Width of the distribution in Y (variance).
			 * @return Updates the content of bufferM.\n
			 */
			BDLL void gauss2D( Ipp64f gain, Ipp64f x0, Ipp64f sigmaX, 
							 Ipp64f y0, Ipp64f sigmaY );

			/** \brief	Normalized 2D Gaussian function.*/
			/** Gauss2Dnorm Function. 
			 * Generates a normalized gaussian distribution in two dimensions 
			 * with its maximum at (x0,y0) controlled by 'gain' and variance 
			 * by 'sigmaX' and 'sigmaY' respectively.
			 * This functions is used to create external stimuli in 2D.
			 * \f{equation}{ 
				 * f(x,y) = C\;exp\left[ \frac{-(x-x0)^2}{2\sigma_x^2} +
				 * \frac{-(y-y0)^2}{2\sigma_y^2} \right] //
				 * f(x,y) = \frac{f(x,y)}{sum{sum{f(x,y)}_y}_x}
			 * \f}
			 * @param gain Amplitude for the gaussian function.
			 * @param x0 Location of the peak in X (mean).
			 * @param sigmaX Width of the distribution in X (variance).
			 * @param y0 Location of the peak in Y (mean).
			 * @param sigmaY Width of the distribution in Y (variance).
			 * @return Updates the content of bufferM.\n
			 */
			BDLL void gauss2Dnorm( Ipp64f gain, Ipp64f x0, Ipp64f sigmaX, 
												Ipp64f y0, Ipp64f sigmaY );

			/** \brief	Normalized Gaussian function for ridge creation.*/
			/** Gauss2D Function for ridge creation. 
			 * Applies a gaussian filter to a 2D vector in a single dimension.
			 * This functions is used to create external stimuli in 2D but in
			 * single dimension.
			 * @param gain Amplitude for the gaussian function.
			 * @param pos Location of the ridge (mean).
			 * @param sigma Width of the distribution (variance).
			 * @param dim 1: creates a ridge in X dimension. 0: creates a ridge
			 * in Y dimension.
			 * @return Updates the content of bufferM.\n
			 */
			BDLL void gauss2Dridge( Ipp64f gain, Ipp64f pos, 
							 Ipp64f sigma, int dim );

			/** \brief	3D Gaussian function.*/
			/** Gauss3D Function. 
			 * Generates gaussian distribution in three dimensions with its 
			 * maximum at (x0,y0,z0) controlled by 'gain' and variance 
			 * by 'sigmaX', 'sigmaY' and 'sigmaZ' respectively.
			 * This functions is used to create external stimuli in 2D.
			 * \f{equation}{ 
				 * f(x,y) = C\;exp\left[ \frac{-(x-x')^2}{2\sigma_x^2} +
				 * \frac{-(y-y')^2}{2\sigma_y^2} + \frac{-(z-z')^2}{2\sigma_z^2} 
				 * \right] 
			 * \f}
			 * @param gain Amplitude for the gaussian function.
			 * @param x0 Location of the peak in X (mean).
			 * @param sigmaX Width of the distribution in X (variance).
			 * @param y0 Location of the peak in Y (mean).
			 * @param sigmaY Width of the distribution in Y (variance).
			 * @param z0 Location of the peak in Z (mean).
			 * @param sigmaZ Width of the distribution in Z (variance).
			 * @return Updates the content of bufferM.\n
			 */
			BDLL void gauss3D( Ipp64f gain, Ipp64f x0, Ipp64f sigmaX, Ipp64f y0, 
										Ipp64f sigmaY, Ipp64f z0, Ipp64f sigmaZ );

			/** \brief	Normalized 3D Gaussian function.*/
			/** Gauss3Dnorm Function. 
			 * Generates a normalized gaussian distribution in three dimensions 
			 * with its maximum at (x0,y0,z0) controlled by 'gain' and variance 
			 * by 'sigmaX', 'sigmaY' and 'sigmaZ' respectively.
			 * This functions is used to create external stimuli in 2D.
			 * \f{equation}{ 
				 * f(x,y) = C\;exp\left[ \frac{-(x-x0)^2}{2\sigma_x^2} +
				 * \frac{-(y-y0)^2}{2\sigma_y^2} + \frac{-(z-z0)^2}{2\sigma_z^2}
				 * \right] //
				 * f(x,y) = \frac{f(x,y,z)}{sum{sum{sum{f(x,y,z)}_y}_x}_z}
			 * \f}
			 * @param gain Amplitude for the gaussian function.
			 * @param x0 Location of the peak in X (mean).
			 * @param sigmaX Width of the distribution in X (variance).
			 * @param y0 Location of the peak in Y (mean).
			 * @param sigmaY Width of the distribution in Y (variance).
			 * @param z0 Location of the peak in Z (mean).
			 * @param sigmaZ Width of the distribution in Z (variance).
			 * @return Updates the content of bufferL.\n
			 */
			BDLL void gauss3Dnorm( Ipp64f gain, Ipp64f x0, Ipp64f sigmaX, Ipp64f y0, 
												Ipp64f sigmaY, Ipp64f z0, Ipp64f sigmaZ );

	};
}

#endif  // BADTOOLS_H

